// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2003 University of Colorado
//                2011 Antonio Carzaniga
//
//  Siena is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//  
//  Siena is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with Siena.  If not, see <http://www.gnu.org/licenses/>.
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cassert>

#include <siena/allocator.h>
#include <siena/atable.h>

namespace siena {

class Attribute;

// 
// This is an implementation of an attribute table based on a ternary
// search trie (TST).  I am not too smart about the balancing of the
// TST, but I do balance each per-character BST that composes the TST.
//
class ATableNode {
public:
    ATableNode(int xc) 
    : data(0), c(xc), left(0), middle(0), right(0) {};

    Attribute *		data;
    int			c;
    ATableNode *	left;
    ATableNode *	middle;
    ATableNode *	right;
};

void ATable::clear() {
    //
    // I don't actually deallocate stuff here since I'm assuming that
    // ftmemory will be deallocated everything at once
    //
    for(int i = 0; i < 256; ++i) roots[i] = 0;
}

//
// these two algorithms are taken more or less directly from
// R. Sedgewick's "Algorithms in C" 3rd Ed. pp 638--639.
//
// Two differences:
//
// 1) I use an aray of 256 root pointers indexed by the first
// character of the string, instead of a single root.  I guess this
// modification doesn't allow me to store (and retrieve) the empty
// string, which is fine for my application.
//
// 2) These version are not recursive.
//
Attribute ** ATable::insert(const char * s, const char * end, 
			  FTAllocator & ftmemory) {
    assert(s != end);
    ATableNode * pp = 0;
    ATableNode ** p = &(roots[static_cast<unsigned char>(*s++)]);
    while (*p != 0) {
	pp = *p;
	if (s == end) {
	    if (pp->c == EndChar) return &(pp->data);
            p = &(pp->left);
	} else if (*s == pp->c) {
	    ++s;
            p = &(pp->middle);
        } else if (*s < pp->c) {
            p = &(pp->left);
	} else {
            p = &(pp->right);
	}
    }

    for (;;) {
	if (s == end) {
	    *p = new (ftmemory)ATableNode(EndChar);
	    return &((*p)->data);
	}
	*p = new (ftmemory)ATableNode(*s);
	pp = *p;
	++s;
        p = &(pp->middle);
    }
}

const Attribute * ATable::find(register const char * s, 
			     const char * end) const {
    assert(s != end);
    register const ATableNode * p = roots[static_cast<unsigned char>(*s++)];
    while (p != 0) {
	if (s == end) {
	    if (p->c == EndChar) return p->data;
            p = p->left;
	} else if (*s < p->c) {
            p = p->left;
	} else if (*s == p->c)  {
	    ++s;
            p = p->middle;
        } else {
            p = p->right;
	}
    }
    return 0;
}

static void right_rotate(ATableNode ** pp) {
    // assume *pp != 0 and (*pp)->left != 0
    ATableNode * p_left = (*pp)->left;
    ATableNode * p_left_right = p_left->right;
    (*pp)->left = p_left_right;
    p_left->right = *pp;
    *pp = p_left;
}

static unsigned int tree_to_vine(ATableNode ** rootp) {
    unsigned int size = 0;
    while(*rootp != 0) {
	while((*rootp)->left != 0) 
	    right_rotate(rootp);
	++size;
	rootp = &((*rootp)->right);
    }
    return size;
}

static void left_rotate(ATableNode ** pp) {
    // assume *pp != 0 and (*pp)->right != 0
    ATableNode * p_right = (*pp)->right;
    ATableNode * p_right_left = p_right->left;
    (*pp)->right = p_right_left;
    p_right->left = *pp;
    *pp = p_right;
}

static void compress(ATableNode ** pp, unsigned int count) {
    while (count > 0) {
	left_rotate(pp);
	pp = &((*pp)->right);
	--count;
    }
}

static unsigned int sup_power_of_two(unsigned int x) {
    // returns the largest power of 2 that is less than or equal to x
    unsigned int y;
    for(;;) {
	y = x & (x - 1);  // clear the least significant bit set
	if (y == 0)
	    break;
	x = y;
    }
    return x;
}

static void vine_to_tree(ATableNode ** rootp, unsigned int size) {
    unsigned int leaf_count = size + 1 - sup_power_of_two(size + 1);
    compress(rootp, leaf_count);
    size -= leaf_count;
    while(size > 1) 
	compress(rootp, size /= 2);
}

//
// This is the BST balancing algorithm described in "Tree Rebalancing
// in Optimal Time and Space" by Q. F. Stout and B. L. Warren,
// Communications of the ACM (CACM) Volume 29 Issue 9, Sept. 1986
//
static void balance_bst(ATableNode ** rootp) {
    unsigned int size;
    size = tree_to_vine(rootp);
    vine_to_tree(rootp, size);
}

static void balance_tst(ATableNode ** rootp);

static void apply_balance_tst(ATableNode * root) {
    if (root->middle)
	balance_tst(&(root->middle));
    if (root->left)
	apply_balance_tst(root->left);
    if (root->right)
	apply_balance_tst(root->right);
}

static void balance_tst(ATableNode ** rootp) {
    // assumes *rootp != 0
    balance_bst(rootp);
    apply_balance_tst(*rootp);
}

void ATable::consolidate() {
    for(int i = 0; i < 256; ++i) 
	if (roots[i])
	    balance_tst(&(roots[i]));
}

} // end namespace siena
