// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2013 Antonio Carzaniga
//
//  Siena is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//  
//  Siena is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with Siena.  If not, see <http://www.gnu.org/licenses/>.
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cstring>
#include <set>
#include <vector>

#include <siena/forwarding.h>
#include <siena/allocator.h>
#include <siena/btrietable.h>

#include "bset_encoding.h"
#include "mmhwrapper.h"

#include "timers.h"

/** 
 *  This file contains the implementation of BTrieTable.
 *
 *  BTrieTable is a forwarding table based on Bloom filters, where
 *  each interface is associated with a predicate consisting of a set
 *  of Bloom filters, each representing a conjunction of constraints,
 *  or in some cases "tags".  Thus the BTrieTable is essentially a
 *  list of BTriePredicate objects, each representing the association
 *  of a predicate to an interface.  
 * 
 *  A Bloom filter, representing a filter, is here represented as a
 *  set of its constituent 1-bits, that is a set of the positions of
 *  its bits set to 1.  So, for example, 00100010110 would be
 *  represented as the set {2,6,8,9} (first position is 0).
 * 
 *  Each predicate, which again consists of a set of Bloom filters, is
 *  then represented as a binary trie.  Each node in the trie
 *  represents an element in a Bloom filter, that is, the position of
 *  a 1-bit.  The elements are in ascending order in the trie, so a
 *  sub-trie rooted at node T with value T.pos represents all the
 *  Bloom filters that contain all the elements between the root and
 *  T, which are less than T.pos, and that may contain other elements
 *  greater than T.pos.  More specifically, a node T has two children
 *  T.t and T.f pointing to the subtrees that represet the Bloom
 *  filters that share the prefix of T (from the root to T) and,
 *  respectively, the Bloom filters that contain T.pos and those that
 *  do not contain T.pos.
 *
 *  The central algorithm of BTrieTable is a subset test, implemented
 *  in suffix_contains_subset.  This test takes a set F of positions
 *  and a binary trie T, and checks whether one of the filters
 *  represented in T is a subset of F.  This algorithm is used in
 *  building the trie representing a predicate in order to avoid
 *  inserting redundant filters (i.e., filters that cover another
 *  filter alredy in the trie).  The same algorithm is also used in the
 *  matching algorithm, since that is essentially a subset search.
 *
 **/

namespace siena {

/** @brief Width of the bloom filters (a.k.a. M)
 **/
static const unsigned int BSize = CONFIG_BLOOM_FILTER_SIZE;


/** @brief representation of a Bloom filters as a set of positions.
 **/
class pos_set_filter {
public:
    typedef unsigned int pos_t;
    typedef std::set<pos_t>::iterator iterator;
    typedef std::set<pos_t>::const_iterator const_iterator;

    static const unsigned int FILTER_SIZE = BSize;

private:
    std::set<pos_t> elements;

public:
    pos_set_filter * next;

    void set_bit(pos_t pos) {
	elements.insert(pos);
    }

    static unsigned int size() { 
	return BSize;
    }

    const_iterator begin() const {
	return elements.begin();
    }

    iterator begin() {
	return elements.begin();
    }

    const_iterator end() const {
	return elements.end();
    }

    iterator end() {
	return elements.end();
    }

    int count() const {
	return elements.size();
    }

    pos_set_filter(): elements() { }

    pos_set_filter(const std::string & s) throw(int);
    pos_set_filter & operator=(const std::string & s) throw(int);
};

class btrie_node {
public:
    pos_set_filter::pos_t pos;
    btrie_node * t;
    btrie_node * f;

    btrie_node(pos_set_filter::pos_t p): pos(p), t(0), f(0) {};
};

static bool suffix_contains_subset(bool prefix_is_good, 
				   const btrie_node * n, 
				   pos_set_filter::const_iterator fi, 
				   pos_set_filter::const_iterator end) {
    // 
    // two alternative implementations of precisely the same
    // algorithm.  The second one is the one that illustrates the
    // algorithm more clearly, and is written in completely recursive
    // form.  The first one, which is the one we actually use, simply
    // transforms the tail-recursive calls into simple iterations.
    //
#ifndef TAIL_RECURSIVE_IS_ITERATIVE
#define TAIL_RECURSIVE_IS_ITERATIVE
#endif

#ifdef TAIL_RECURSIVE_IS_ITERATIVE
    while(n != 0) {
	if (fi == end) {
	    return false;
	} else if (n->pos > *fi) {
	    ++fi; 
	} else if (n->pos < *fi) {
	    prefix_is_good = false;
	    n = n->f;
	} else if (suffix_contains_subset(false, n->f, ++fi, end)) {
	    return true;
	} else {
	    prefix_is_good = true;
	    n = n->t;
	}
    }
    return prefix_is_good;
#else
     if (n == 0)
 	return prefix_is_good;
     if (fi == end)
 	return false;
     if (n->pos > *fi) 
 	return suffix_contains_subset(prefix_is_good, n, ++fi, end);
     if (n->pos < *fi)
 	return suffix_contains_subset(false, n->f, fi, end);
     // n->pos == *fi
     ++fi;
     return suffix_contains_subset(false, n->f, fi, end)
	 || suffix_contains_subset(true, n->t, fi, end);
#endif
}

/** @brief a predicate in a BTrieTable.
 *
 *  This is an extremely simple data structure.  The predicate is
 *  represented as a linked list of filters.  Thus, BTriePredicate is
 *  a link in such a list and refers directly to its interface id and
 *  the trie reprenting the set of filters that make up the predicate.
 **/
struct BTriePredicate {
    btrie_node * root;
    ifid_t id;
    BTriePredicate * next;

public:
    BTriePredicate(BTriePredicate *n, ifid_t i) 
	: root(0), id(i), next(n) {}

    ~BTriePredicate(); 

    bool contains_subset(pos_set_filter::const_iterator begin, 
			 pos_set_filter::const_iterator end) const {
	return suffix_contains_subset(false, root, begin, end);
    }

    void add_filter(pos_set_filter::const_iterator begin, 
		    pos_set_filter::const_iterator end,
		    FTAllocator & memory);
};

void BTriePredicate::add_filter(pos_set_filter::const_iterator fi, 
				pos_set_filter::const_iterator end,
				FTAllocator & memory) {
    btrie_node ** np = &root;

    while(fi != end) {
	if (*np == 0) {
	    *np = new (memory)btrie_node(*fi);
	    np = &((*np)->t);
	    ++fi;
	} else if ((*np)->pos < *fi) {
	    np = &((*np)->f);
	} else if ((*np)->pos > *fi) {
	    btrie_node * tmp = *np;
	    *np = new (memory)btrie_node(*fi);
	    (*np)->f = tmp;
	    np = &((*np)->t);
	    ++fi;
	} else { // (*np)->pos == *fi
	    np = &((*np)->t);
	    ++fi;
	}
    }
}

BTrieTable::BTrieTable() : plist(0) {}
BTrieTable::~BTrieTable() {}

void BTrieTable::ifconfig(if_t id, const predicate & p) {

    TIMER_PUSH(ifconfig_timer);

    predicate::iterator * pi = p.first();

    if (pi) {
	BTriePredicate * p = new (memory)BTriePredicate(plist, id);
	plist = p;

	pos_set_filter * fv[BSize];
	memset(fv,0,sizeof(fv));
	do {
	    pos_set_filter * b = new pos_set_filter();

	    TIMER_PUSH(bloom_encoding_timer);

	    encode_filter(*b, pi);

	    TIMER_POP();

	    b->next = fv[b->count()];
	    fv[b->count()] = b;
	} while(pi->next());
	delete(pi);

	for (unsigned int i = 0; i < BSize; ++i) {
	    pos_set_filter * b = fv[i]; 
	    pos_set_filter * tmp;
	    while (b != 0) {
		if (!p->contains_subset(b->begin(), b->end()))
		    p->add_filter(b->begin(), b->end(), memory);
		tmp = b;
		b = b->next;
		delete(tmp);
	    }
	}
    }
    TIMER_POP();
}

void BTrieTable::match(const message & m, MatchHandler & h) const {

    BTriePredicate * p = plist;
    if (!p) return;

    TIMER_PUSH(bloom_encoding_timer);

    pos_set_filter b;
    encode_message(b, &m);

    TIMER_POP();

    TIMER_PUSH(match_timer);

    do {
	if (p->contains_subset(b.begin(), b.end())) {

	    TIMER_PUSH(forward_timer);

	    bool output_result = h.output(p->id);

	    TIMER_POP();

	    if (output_result) {
		TIMER_POP();
		return;
	    }
	}
    } while ((p = p->next));

    TIMER_POP();
}

void BTrieTable::match(const message & m, MatchMessageHandler & p) const {
    MMHWrapper w(p, m);
    match(m, w);
}

void BTrieTable::clear() {
    memory.clear();
    plist = 0;
}
void BTrieTable::clear_recycle() {
    memory.recycle();
    plist = 0;
}

size_t BTrieTable::allocated_bytesize() const {
    return memory.allocated_size();
}

size_t BTrieTable::bytesize() const {
    return memory.size();
}

} // end namespace siena
